
{*******************************************************}
{                                                       }
{       InterBase Connections Service Version 1.0       }
{                                                       }
{                                                       }
{          Copyright (c) 1999-2000 Vadim Crits          }
{                                                       }
{*******************************************************}

unit Main;

{$O-}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, SvcMgr, Dialogs,
  Registry, WinSock;

type
  TInterBaseConnections = class(TService)
    procedure ServiceCreate(Sender: TObject);
    procedure ServiceStart(Sender: TService; var Started: Boolean);
    procedure ServicePause(Sender: TService; var Paused: Boolean);
    procedure ServiceContinue(Sender: TService; var Continued: Boolean);
    procedure ServiceStop(Sender: TService; var Stopped: Boolean);
    procedure ServiceShutdown(Sender: TService);
  private
    { Private declarations }
  public
    { Public declarations }
    function GetServiceController: TServiceController; override;
  end;

const
  IB_KEY     = '\SOFTWARE\InterBase Corp\InterBase\CurrentVersion';
  IB_LOG     = 'interbase.log';
  S_FORMAT   = '%s (%s)'#9'%s'#13#10#9'%s: IP address - %s, Hostname - %s'#13#10#13#10#13#10;
  S_DTFORMAT = 'ddd mmm dd hh:nn:ss yyyy';
  ANY_SIZE   = 511;
  TIMEOUT    = 1000;

{ Specifies the state of the TCP connection }

  MIB_TCP_STATE_CLOSED     = 1;
  MIB_TCP_STATE_LISTEN     = 2;
  MIB_TCP_STATE_SYN_SENT   = 3;
  MIB_TCP_STATE_SYN_RCVD   = 4;
  MIB_TCP_STATE_ESTAB      = 5;
  MIB_TCP_STATE_FIN_WAIT1  = 6;
  MIB_TCP_STATE_FIN_WAIT2  = 7;
  MIB_TCP_STATE_CLOSE_WAIT = 8;
  MIB_TCP_STATE_CLOSING    = 9;
  MIB_TCP_STATE_LAST_ACK   = 10;
  MIB_TCP_STATE_TIME_WAIT  = 11;
  MIB_TCP_STATE_DELETE_TCB = 12;


type
  PMibTcpRow = ^TMibTcpRow;
  TMibTcpRow = packed record
    dwState: DWORD;
    dwLocalAddr: DWORD;
    dwLocalPort: DWORD;
    dwRemoteAddr: DWORD;
    dwRemotePort: DWORD;
  end;

  PMibTcpTable = ^TMibTcpTable;
  TMibTcpTable = packed record
    dwNumEntries: DWORD;
    table: array[0..ANY_SIZE] of TMibTcpRow;
  end;

  PIBConRow = ^TIBConRow;
  TIBConRow = packed record
    dwState: DWORD;
    dwRemoteAddr: DWORD;
    dwRemotePort: DWORD;
    HostName: string;
  end;

  PIBConTable = ^TIBConTable;
  TIBConTable = packed record
    dwNumEntries: DWORD;
    table: array[0..ANY_SIZE] of TIBConRow;
  end;

  TIBConThread = class(TThread)
  private
    FIBConTable: TIBConTable;
  protected
    procedure Execute; override;
  end;

  TIBFindThread = class(TThread)
  protected
    procedure Execute; override;
  end;

function GetTcpTable(pTcpTable: PMibTcpTable; pdwSize: PDWORD;
  bOrder: BOOL): DWORD; stdcall; external 'iphlpapi.dll'  name 'GetTcpTable';

var
  InterBaseConnections: TInterBaseConnections;

implementation

{$R *.DFM}

var
  IBConThread: TIBConThread;
  IBFindThread: TIBFindThread;
  IBRootDirectory: string;
  ComputerName: array[0..MAX_COMPUTERNAME_LENGTH - 1] of Char;

procedure ServiceController(CtrlCode: DWord); stdcall;
begin
     InterBaseConnections.Controller(CtrlCode);
end;

function TInterBaseConnections.GetServiceController: TServiceController;
begin
     Result := ServiceController;
end;

procedure TInterBaseConnections.ServiceCreate(Sender: TObject);
var
   Reg: TRegistry;
   WSAData: TWSAData;
   dwSize: DWORD;
begin
     Reg := TRegistry.Create;
     try
        with Reg do
        begin
             RootKey := HKEY_LOCAL_MACHINE;
             if KeyExists(IB_KEY) then
             begin
                  OpenKey(IB_KEY, False);
                  IBRootDirectory := ReadString('RootDirectory');
                  CloseKey;
             end
             else
             begin
                  MessageBox(0, 'InterBase server not found.', PChar(DisplayName),
                             MB_OK or MB_ICONERROR);
                  Halt(1);
             end;
        end;
     finally
        Reg.Free;
     end;
     SetThreadLocale($0409);
     GetFormatSettings;
     dwSize := SizeOf(ComputerName);
     GetComputerName(ComputerName, dwSize);
     WSAStartup($0101, WSAData);
     SetPriorityClass(GetCurrentProcess, IDLE_PRIORITY_CLASS);
end;

procedure TInterBaseConnections.ServiceStart(Sender: TService;
  var Started: Boolean);
begin
     IBConThread := TIBConThread.Create(False);
     IBFindThread := TIBFindThread.Create(False);
     Started := True;
end;

procedure TInterBaseConnections.ServicePause(Sender: TService;
  var Paused: Boolean);
begin
     IBConThread.Suspend;
     IBFindThread.Suspend;
     Paused := True;
end;

procedure TInterBaseConnections.ServiceContinue(Sender: TService;
  var Continued: Boolean);
begin
     IBConThread.Resume;
     IBFindThread.Resume;
     Continued := True;
end;

procedure TInterBaseConnections.ServiceStop(Sender: TService;
  var Stopped: Boolean);
begin
     IBConThread.Terminate;
     IBFindThread.Terminate;
     Stopped := True;
end;

procedure TInterBaseConnections.ServiceShutdown(Sender: TService);
begin
     WSACleanup;
end;

procedure TIBConThread.Execute;

   function FormatIP(Value: DWORD): string;
   begin
        Result := Format('%d.%d.%d.%d', [Value and $FF,
                                         Value shr 8 and $FF,
                                         Value shr 16 and $FF,
                                         Value shr 24 and $FF]);
   end;

var
   TcpTable: TMibTcpTable;
   HostEnt: PHostEnt;
   dwSize, dwResult, i, j: DWORD;
   NewCon: Boolean;
   FileHandle: Integer;
   s: string;
begin
     while not Terminated do
     begin
          dwSize := SizeOf(TMibTcpTable);
          FillChar(TcpTable, dwSize, 0);
          dwResult := GetTcpTable(@TcpTable, @dwSize, True);
          if dwResult <> NO_ERROR then Continue;
          with TcpTable do for i := 0 to dwNumEntries - 1 do
          if not ((table[i].dwState = MIB_TCP_STATE_ESTAB) and (table[i].dwLocalPort = 59915)) then
          begin
               FillChar(table[i], SizeOf(TMibTcpRow), 0);
               Dec(dwNumEntries);
          end;
          with FIBConTable do if dwNumEntries <> 0 then for i := 0 to High(table) do
          if table[i].dwState <> 0 then table[i].dwState := MIB_TCP_STATE_CLOSED;
          with TcpTable do if dwNumEntries <> 0 then
          for i := 0 to High(table) do if table[i].dwState <> 0 then
          begin
               NewCon := True;
               for j := 0 to High(FIBConTable.table) do
               if table[i].dwRemotePort = FIBConTable.table[j].dwRemotePort then
               begin
                    NewCon := False;
                    FIBConTable.table[j].dwState := table[i].dwState;
                    Break;
               end;
               if NewCon then for j := 0 to High(FIBConTable.table) do
               if FIBConTable.table[j].dwState = 0 then
               begin
                    HostEnt := gethostbyaddr(@table[i].dwRemoteAddr, 4, PF_INET);
                    FIBConTable.table[j].dwState := table[i].dwState;
                    FIBConTable.table[j].dwRemoteAddr := table[i].dwRemoteAddr;
                    FIBConTable.table[j].dwRemotePort := table[i].dwRemotePort;
                    if HostEnt <> nil then
                       FIBConTable.table[j].HostName := HostEnt^.h_name
                    else
                       FIBConTable.table[j].HostName := 'UNKNOWN';
                    Inc(FIBConTable.dwNumEntries);
                    FileHandle := FileOpen(IBRootDirectory + IB_LOG, fmOpenReadWrite or fmShareDenyNone);
                    if FileHandle <> -1 then
                    begin
                         s := Format(S_FORMAT, [ComputerName,
                                                InterBaseConnections.DisplayName,
                                                FormatDateTime(S_DTFORMAT, Now),
                                                'Connect',
                                                FormatIP(FIBConTable.table[j].dwRemoteAddr),
                                                FIBConTable.table[j].HostName]);
                         FileSeek(FileHandle, 0, 2);
                         FileWrite(FileHandle, PChar(s)^, Length(s));
                         FileClose(FileHandle);
                    end;
                    Break;
               end;
          end;
          with FIBConTable do if dwNumEntries <> 0 then
          for i := 0 to High(table) do if table[i].dwState = MIB_TCP_STATE_CLOSED then
          begin
               FileHandle := FileOpen(IBRootDirectory + IB_LOG, fmOpenReadWrite or fmShareDenyNone);
               if FileHandle <> -1 then
               begin
                    s := Format(S_FORMAT, [ComputerName,
                                           InterBaseConnections.DisplayName,
                                           FormatDateTime(S_DTFORMAT, Now),
                                           'Disconnect',
                                           FormatIP(table[i].dwRemoteAddr),
                                           table[i].HostName]);
                    FileSeek(FileHandle, 0, 2);
                    FileWrite(FileHandle, PChar(s)^, Length(s));
                    FileClose(FileHandle);
               end;
               FillChar(table[i], SizeOf(TIBConRow), 0);
               Dec(dwNumEntries)
          end;
          Sleep(TIMEOUT);
     end;
end;

procedure TIBFindThread.Execute;
begin
     while not Terminated do
     begin
          if FindWindow('IB_Server', 'InterBase Server') = 0 then
          begin
               if not IBConThread.Suspended then IBConThread.Suspend;
          end
          else if IBConThread.Suspended then IBConThread.Resume;
          Sleep(TIMEOUT * 12);
     end;
end;

end.
